unit System.Windows.Forms.NullableDateTimePicker;

//This code is translated from the article at
//http://www.codeproject.com/cs/miscctrl/NullableDateTimePicker.asp
//and lets a DateTimePicker support null values

interface

uses
  System.Windows.Forms;

type
  DateTimePicker = class(System.Windows.Forms.DateTimePicker)
  private
    procedure set_FormatAsString(const Value: string);
  private
    // true, when no date shall be displayed (empty DateTimePicker)
    _isNull: Boolean;
    // If _isNull = True, this value is shown in the DTP
    _nullValue: String;
    // The format of the DateTimePicker control
     _format: DateTimePickerFormat;
    // The custom format of the DateTimePicker control
    _customFormat: String;
    // The format of the DateTimePicker control as string
    _formatAsString: String;
    procedure SetFormat;
    procedure SetToDateTimeValue;
    procedure SetToNullValue;
    property FormatAsString: string read _formatAsString write set_FormatAsString;
  strict protected
    procedure WndProc(var m: Message); override;
    procedure OnKeyUp(e: KeyEventArgs); override;
  public
    procedure set_Format(const Value: DateTimePickerFormat);
    function get_Value: &Object;
    procedure set_Value(const Value: &Object);
  public
    constructor Create;
    property CustomFormat: String read _CustomFormat write _CustomFormat;
    property Format: DateTimePickerFormat read _Format write set_Format;
    property NullValue: String read _nullValue write _nullValue;
    property Value: &Object read get_Value write set_Value;
  end;

implementation

uses
  System.Globalization, System.Threading,
  System.Runtime.InteropServices;

{ DateTimePicker }

constructor DateTimePicker.Create;
begin
  inherited;
  inherited Format := DateTimePickerFormat.Custom;
  NullValue := ' ';
  Format := DateTimePickerFormat.Long
end;

function DateTimePicker.get_Value: &Object;
begin
  if _isNull then
    Result := nil
  else
    Result := inherited Value;
end;

procedure DateTimePicker.OnKeyUp(e: KeyEventArgs);
begin
  if (e.KeyCode = Keys.Delete) then
  begin
    Value := NullValue;
    OnValueChanged(EventArgs.Empty);
  end;
  inherited;
end;

procedure DateTimePicker.SetFormat;
var
  ci: CultureInfo;
  dtf: DateTimeFormatInfo;
begin
  ci := Thread.CurrentThread.CurrentCulture;
  dtf := ci.DateTimeFormat;
  case _format of
    DateTimePickerFormat.Long:
      FormatAsString := dtf.LongDatePattern;
    DateTimePickerFormat.Short:
      FormatAsString := dtf.ShortDatePattern;
    DateTimePickerFormat.Time:
      FormatAsString := dtf.ShortTimePattern;
    DateTimePickerFormat.Custom:
      FormatAsString := CustomFormat;
  end
end;

procedure DateTimePicker.SetToDateTimeValue;
begin
  if _isNull then
  begin
    SetFormat;
    _isNull := False;
    inherited OnValueChanged(EventArgs.Create)
  end;
end;

procedure DateTimePicker.SetToNullValue;
begin
  _isNull := True;
  if (_nullValue = nil) or (_nullValue = System.String.Empty) then
    inherited CustomFormat := ' '
  else
    inherited CustomFormat := '''' + NullValue + ''''
end;

procedure DateTimePicker.set_Format(const Value: DateTimePickerFormat);
begin
  _format := value;
  if not _isNull then
    SetFormat;
  OnFormatChanged(EventArgs.Empty);
end;

procedure DateTimePicker.set_FormatAsString(const Value: string);
begin
  _formatAsString := value;
  inherited CustomFormat := value;
end;

procedure DateTimePicker.set_Value(const Value: &Object);
begin
  if (value = nil) or (value = DBNull.Value) then
    SetToNullValue
  else
  begin
    SetToDateTimeValue;
    inherited Value := DateTime(value)
  end
end;

type
  [StructLayout(LayoutKind.Sequential)]
  NMHDR = record
    HwndFrom: IntPtr;
    IdFrom: Integer;
    Code: Integer;
  end;

procedure DateTimePicker.WndProc(var m: Message);
var
  nm: NMHDR;
begin
  if m.Msg = $4e then  // WM_NOTIFY
  begin
    nm := NMHDR(m.GetLParam(TypeOf(NMHDR)));
    if (nm.Code = -746) or (nm.Code = -722) then // DTN_CLOSEUP || DTN_?
      SetToDateTimeValue
  end;
  inherited;
end;

end.
